#ifdef HAMMER_LLVM_BACKEND
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wpedantic"
#include <llvm-c/Core.h>
#pragma GCC diagnostic pop
#include "../backends/llvm/llvm.h"
#endif

#include "parser_internal.h"

typedef struct {
  const HParser *p;
  int64_t lower;
  int64_t upper;
} HRange;

static HParseResult* parse_int_range(void *env, HParseState *state) {
  HRange *r_env = (HRange*)env;
  HParseResult *ret = h_do_parse(r_env->p, state);
  if (!ret || !ret->ast)
    return NULL;
  switch(ret->ast->token_type) {
  case TT_SINT:
    if (r_env->lower <= ret->ast->sint && r_env->upper >= ret->ast->sint)
      return ret;
    else
      return NULL;
  case TT_UINT:
    if ((uint64_t)r_env->lower <= ret->ast->uint && (uint64_t)r_env->upper >= ret->ast->uint)
      return ret;
    else
      return NULL;
  default:
    return NULL;
  }
}

void gen_int_range(HAllocator *mm__, HCFStack *stk__, uint64_t low, uint64_t high, uint8_t bytes) {
  /* Possible FIXME: TallerThanMe */
  if (1 == bytes) {
    HCharset cs = new_charset(mm__);
    for (uint64_t i=low; i<=high; ++i) {
      charset_set(cs, i, 1);
    }
    HCFS_ADD_CHARSET(cs);
  }
  else if (1 < bytes) {
    uint8_t low_head, hi_head;
    low_head = ((low >> (8*(bytes - 1))) & 0xFF);
    hi_head = ((high >> (8*(bytes - 1))) & 0xFF);
    if (low_head != hi_head) {
      HCFS_BEGIN_CHOICE() {
	HCFS_BEGIN_SEQ() {
	  HCFS_ADD_CHAR(low_head);
	  gen_int_range(mm__, stk__, low & ((1 << (8 * (bytes - 1))) - 1), ((1 << (8*(bytes-1)))-1), bytes-1);
	} HCFS_END_SEQ();
	HCFS_BEGIN_SEQ() {
	  HCharset hd = new_charset(mm__);
	  HCharset rest = new_charset(mm__);
	  for (int i = 0; i < 256; i++) {
	    charset_set(hd, i, (i > low_head && i < hi_head));
	    charset_set(rest, i, 1);
	  }
	  HCFS_ADD_CHARSET(hd);
	  for (int i = 2; i < bytes; i++)
	    HCFS_ADD_CHARSET(rest);
	} HCFS_END_SEQ();
	HCFS_BEGIN_SEQ() {
	  HCFS_ADD_CHAR(hi_head);
	  gen_int_range(mm__, stk__, 0, high & ((1 << (8 * (bytes - 1))) - 1), bytes-1);
	} HCFS_END_SEQ();
      } HCFS_END_CHOICE();
    } else {
      // TODO: find a way to merge this with the higher-up SEQ
      HCFS_BEGIN_CHOICE() {
	HCFS_BEGIN_SEQ() {
	  HCFS_ADD_CHAR(low_head);
	  gen_int_range(mm__, stk__,
			low & ((1 << (8 * (bytes - 1))) - 1),
			high & ((1 << (8 * (bytes - 1))) - 1),
			bytes - 1);
	} HCFS_END_SEQ();
      } HCFS_END_CHOICE();
    }
  }
}

struct bits_env {
  uint8_t length;
  uint8_t signedp;
};

static void desugar_int_range(HAllocator *mm__, HCFStack *stk__, void *env) {
  HRange *r = (HRange*)env;
  struct bits_env* be = (struct bits_env*)r->p->env;
  uint8_t bytes = be->length / 8;
  gen_int_range(mm__, stk__, r->lower, r->upper, bytes);
}

bool h_svm_action_validate_int_range(HArena *arena, HSVMContext *ctx, void* env) {
  HRange *r_env = (HRange*)env;
  HParsedToken *head = ctx->stack[ctx->stack_count-1];
  switch (head-> token_type) {
  case TT_SINT: 
    return head->sint >= r_env->lower && head->sint <= r_env->upper;
  case TT_UINT: 
    return head->uint >= (uint64_t)r_env->lower && head->uint <= (uint64_t)r_env->upper;
  default:
    return false;
  }
}

static bool ir_ctrvm(HRVMProg *prog, void *env) {
  HRange *r_env = (HRange*)env;
  
  h_compile_regex(prog, r_env->p);
  h_rvm_insert_insn(prog, RVM_ACTION, h_rvm_create_action(prog, h_svm_action_validate_int_range, env));
  return false;
}

#ifdef HAMMER_LLVM_BACKEND

static bool ir_llvm(HLLVMParserCompileContext *ctxt,
                    const HParser *p,
                    LLVMValueRef *res) {
  /* Build some LLVM IR to parse an intrange */

  /* The allocator */
  HAllocator *mm__ = NULL;
  /* Names */
  H_LLVM_DECLARE_NAME(ir_entry);
  H_LLVM_DECLARE_NAME(ir_check_res);
  H_LLVM_DECLARE_NAME(ir_check_res_ast);
  H_LLVM_DECLARE_NAME(ir_check_token_type);
  H_LLVM_DECLARE_NAME(ir_check_tt_uint);
  H_LLVM_DECLARE_NAME(ir_check_tt_sint);
  H_LLVM_DECLARE_NAME(ir_handle_tt_uint);
  H_LLVM_DECLARE_NAME(ir_handle_tt_sint);
  H_LLVM_DECLARE_NAME(ir_check_bounds);
  H_LLVM_DECLARE_NAME(ir_check_upper_bound);
  H_LLVM_DECLARE_NAME(ir_success);
  H_LLVM_DECLARE_NAME(ir_fail);
  H_LLVM_DECLARE_NAME(ir_end);
  H_LLVM_DECLARE_NAME(ir_inner_res_icmp);
  H_LLVM_DECLARE_NAME(ir_ret_ast_ptr);
  H_LLVM_DECLARE_NAME(ir_ret_ast);
  H_LLVM_DECLARE_NAME(ir_ret_ast_icmp);
  H_LLVM_DECLARE_NAME(ir_token_type_ptr);
  H_LLVM_DECLARE_NAME(ir_token_type);
  H_LLVM_DECLARE_NAME(ir_token_data_ptr);
  H_LLVM_DECLARE_NAME(ir_token_data);
  H_LLVM_DECLARE_NAME(ir_tt_uint_icmp);
  H_LLVM_DECLARE_NAME(ir_tt_sint_icmp);
  H_LLVM_DECLARE_NAME(ir_lower_uint_icmp);
  H_LLVM_DECLARE_NAME(ir_lower_sint_icmp);
  H_LLVM_DECLARE_NAME(ir_lower_icmp);
  H_LLVM_DECLARE_NAME(ir_upper_uint_icmp);
  H_LLVM_DECLARE_NAME(ir_upper_sint_icmp);
  H_LLVM_DECLARE_NAME(ir_upper_icmp);
  H_LLVM_DECLARE_NAME(ir_rv);
  /* The HRange */
  HRange *r = NULL;
  /* Our status */
  bool status = true;
  /* intrange basic blocks */
  LLVMBasicBlockRef entry, check_res, check_res_ast, check_token_type;
  LLVMBasicBlockRef check_tt_uint, check_tt_sint, handle_tt_uint, handle_tt_sint;
  LLVMBasicBlockRef check_bounds, check_upper_bound;
  LLVMBasicBlockRef success, fail, end;
  /* Result of the parser this intrange wraps */
  LLVMValueRef inner_res;
  /* Result of result-not-zero icmp */
  LLVMValueRef inner_res_icmp;
  /* Pointer to AST pointer from inner_res */
  LLVMValueRef ret_ast_ptr;
  /* AST pointer from inner_res */
  LLVMValueRef ret_ast;
  /* Result of res_ast != 0 icmp */
  LLVMValueRef ret_ast_icmp;
  /* Pointer to token type from ret->ast */
  LLVMValueRef token_type_ptr;
  /* Value of token type from ret->ast */
  LLVMValueRef token_type;
  /* Pointer to token data from ret->ast */
  LLVMValueRef token_data_ptr;
  /* Value of token data from ret->ast */
  LLVMValueRef token_data;
  /* Result of comparing token_type to TT_UINT */
  LLVMValueRef tt_uint_icmp;
  /* Result of comparing token_type to TT_SINT */
  LLVMValueRef tt_sint_icmp;
  /* Result of icmp against lower bound */
  LLVMValueRef lower_sint_icmp, lower_uint_icmp, lower_icmp;
  /* Result of icmp against upper bound */
  LLVMValueRef upper_sint_icmp, upper_uint_icmp, upper_icmp;
  /* Inputs to the phis for lower_icmp and upper_icmp */
  LLVMBasicBlockRef lower_icmp_incoming_blocks[2];
  LLVMValueRef lower_icmp_incoming_values[2];
  LLVMBasicBlockRef upper_icmp_incoming_blocks[2];
  LLVMValueRef upper_icmp_incoming_values[2];
  /* Our result */
  LLVMValueRef rv;
  /* Inputs to the phi */
  LLVMBasicBlockRef rv_phi_incoming_blocks[2];
  LLVMValueRef rv_phi_incoming_values[2];
  /* Width of the integer in the token type */
  unsigned int token_int_width;


  if (!res) return false;
  if (!p) return false;
  if (!ctxt) return false;

  /* Unpack the allocator */
  mm__ = ctxt->mm__;

  /* Get the range */
  r = (HRange *)(p->env);
  if (!r) return false;

  /* Set up our basic blocks */
  H_LLVM_COMPUTE_NAME(ir_entry, p);
  entry = LLVMAppendBasicBlock(ctxt->func, ir_entry_name);
  H_LLVM_COMPUTE_NAME(ir_check_res, p);
  check_res = LLVMAppendBasicBlock(ctxt->func, ir_check_res_name);
  H_LLVM_COMPUTE_NAME(ir_check_res_ast, p);
  check_res_ast = LLVMAppendBasicBlock(ctxt->func, ir_check_res_ast_name);
  H_LLVM_COMPUTE_NAME(ir_check_token_type, p);
  check_token_type =
    LLVMAppendBasicBlock(ctxt->func, ir_check_token_type_name);
  H_LLVM_COMPUTE_NAME(ir_check_tt_uint, p);
  check_tt_uint =
    LLVMAppendBasicBlock(ctxt->func, ir_check_tt_uint_name);
  H_LLVM_COMPUTE_NAME(ir_check_tt_sint, p);
  check_tt_sint =
    LLVMAppendBasicBlock(ctxt->func, ir_check_tt_sint_name);
  H_LLVM_COMPUTE_NAME(ir_handle_tt_uint, p);
  handle_tt_uint =
    LLVMAppendBasicBlock(ctxt->func, ir_handle_tt_uint_name);
  H_LLVM_COMPUTE_NAME(ir_handle_tt_sint, p);
  handle_tt_sint =
    LLVMAppendBasicBlock(ctxt->func, ir_handle_tt_sint_name);
  H_LLVM_COMPUTE_NAME(ir_check_bounds, p);
  check_bounds = LLVMAppendBasicBlock(ctxt->func, ir_check_bounds_name);
  H_LLVM_COMPUTE_NAME(ir_check_upper_bound, p);
  check_upper_bound = LLVMAppendBasicBlock(ctxt->func,
      ir_check_upper_bound_name);
  H_LLVM_COMPUTE_NAME(ir_success, p);
  success = LLVMAppendBasicBlock(ctxt->func, ir_success_name);
  H_LLVM_COMPUTE_NAME(ir_fail, p);
  fail = LLVMAppendBasicBlock(ctxt->func, ir_fail_name);
  H_LLVM_COMPUTE_NAME(ir_end, p);
  end = LLVMAppendBasicBlock(ctxt->func, ir_end_name);

  /* Basic block: entry */
  LLVMBuildBr(ctxt->builder, entry);
  LLVMPositionBuilderAtEnd(ctxt->builder, entry);

  /* Try the inner parser */
  if (r->p && r->p->vtable && r->p->vtable->llvm) {
    if (r->p->vtable->llvm(ctxt, r->p, &inner_res)) {
      /* w00t, the inner parser compiled */

      /*
       * Okay, now we should have a pointer to an HParseResult in inner_res,
       * and we need to check that the result is a TT_SINT or a TT_UINT and
       * matches the required range.
       *
       * XXX we could probably look at the type of parser we have under the
       * h_int_range at compile time and build a fast-path common case.
       */

      /* Set up the int_range bb */
      LLVMBuildBr(ctxt->builder, check_res);

      /* Check that the result is not null */
      LLVMPositionBuilderAtEnd(ctxt->builder, check_res);
      H_LLVM_COMPUTE_NAME(ir_inner_res_icmp, p);
      inner_res_icmp = LLVMBuildICmp(ctxt->builder, LLVMIntNE,
          inner_res, LLVMConstNull(ctxt->llvm_parseresultptr),
          ir_inner_res_icmp_name);
      LLVMBuildCondBr(ctxt->builder, inner_res_icmp, check_res_ast, fail);

      /* First, get the AST pointer out */
      LLVMPositionBuilderAtEnd(ctxt->builder, check_res_ast);
      H_LLVM_COMPUTE_NAME(ir_ret_ast_ptr, p);
      ret_ast_ptr = LLVMBuildStructGEP(ctxt->builder, inner_res,
          0, ir_ret_ast_ptr_name);

      /* Deref ret_ast_ptr */
      H_LLVM_COMPUTE_NAME(ir_ret_ast, p);
      ret_ast = LLVMBuildLoad(ctxt->builder, ret_ast_ptr, ir_ret_ast_name);

      /* Check that the AST pointer is not null */
      H_LLVM_COMPUTE_NAME(ir_ret_ast_icmp, p);
      ret_ast_icmp = LLVMBuildICmp(ctxt->builder, LLVMIntNE,
          ret_ast, LLVMConstNull(ctxt->llvm_parsedtokenptr),
          ir_ret_ast_icmp_name);
      LLVMBuildCondBr(ctxt->builder, ret_ast_icmp, check_token_type, fail);

      /* Holly shit, it's a real token! */
      LLVMPositionBuilderAtEnd(ctxt->builder, check_token_type);
      H_LLVM_COMPUTE_NAME(ir_token_type_ptr, p);
      token_type_ptr = LLVMBuildStructGEP(ctxt->builder, ret_ast,
          0, ir_token_type_ptr_name);
      H_LLVM_COMPUTE_NAME(ir_token_type, p);
      token_type =
        LLVMBuildLoad(ctxt->builder, token_type_ptr, ir_token_type_name);
      H_LLVM_COMPUTE_NAME(ir_token_data_ptr, p);
      token_data_ptr = LLVMBuildStructGEP(ctxt->builder, ret_ast,
          1, ir_token_data_ptr_name);
      H_LLVM_COMPUTE_NAME(ir_token_data, p);
      token_data =
        LLVMBuildLoad(ctxt->builder, token_data_ptr, ir_token_data_name);
      /* Now check if we have a TT_UINT */
      LLVMBuildBr(ctxt->builder, check_tt_uint);

      /* Are we doing a TT_UINT? */
      LLVMPositionBuilderAtEnd(ctxt->builder, check_tt_uint);
      H_LLVM_COMPUTE_NAME(ir_tt_uint_icmp, p);
      tt_uint_icmp = LLVMBuildICmp(ctxt->builder, LLVMIntEQ,
          token_type, LLVMConstInt(LLVMInt32Type(), TT_UINT, 0),
          ir_tt_uint_icmp_name);
      LLVMBuildCondBr(ctxt->builder, tt_uint_icmp, handle_tt_uint, check_tt_sint);

      /* Are we doing a TT_SINT? */
      LLVMPositionBuilderAtEnd(ctxt->builder, check_tt_sint);
      H_LLVM_COMPUTE_NAME(ir_tt_sint_icmp, p);
      tt_sint_icmp = LLVMBuildICmp(ctxt->builder, LLVMIntEQ,
          token_type, LLVMConstInt(LLVMInt32Type(), TT_SINT, 0),
          ir_tt_sint_icmp_name);
      LLVMBuildCondBr(ctxt->builder, tt_sint_icmp, handle_tt_sint, fail);

      /*
       * No need to dick around with sext/zext on token_data because
       * it was already done on token creation; see h_llvm_make_tt_suint()
       */

      /* Handle the TT_UINT case */
      LLVMPositionBuilderAtEnd(ctxt->builder, handle_tt_uint);
      H_LLVM_COMPUTE_NAME(ir_lower_uint_icmp, p);
      lower_uint_icmp = LLVMBuildICmp(ctxt->builder, LLVMIntULT,
          token_data,
          LLVMConstInt(LLVMTypeOf(token_data), (uint64_t)(r->lower), 0),
          ir_lower_uint_icmp_name);
      H_LLVM_COMPUTE_NAME(ir_upper_uint_icmp, p);
      upper_uint_icmp = LLVMBuildICmp(ctxt->builder, LLVMIntUGT,
          token_data,
          LLVMConstInt(LLVMTypeOf(token_data), (uint64_t)(r->upper), 0),
          ir_upper_uint_icmp_name);
      LLVMBuildBr(ctxt->builder, check_bounds);

      /* Handle the TT_SINT case */
      LLVMPositionBuilderAtEnd(ctxt->builder, handle_tt_sint);
      H_LLVM_COMPUTE_NAME(ir_lower_sint_icmp, p);
      lower_sint_icmp = LLVMBuildICmp(ctxt->builder, LLVMIntSLT,
          token_data,
          LLVMConstInt(LLVMTypeOf(token_data), (int64_t)(r->lower), 0),
          ir_lower_sint_icmp_name);
      H_LLVM_COMPUTE_NAME(ir_upper_sint_icmp, p);
      upper_sint_icmp = LLVMBuildICmp(ctxt->builder, LLVMIntSGT,
          token_data,
          LLVMConstInt(LLVMTypeOf(token_data), (int64_t)(r->upper), 0),
          ir_upper_sint_icmp_name);
      LLVMBuildBr(ctxt->builder, check_bounds);

      /*
       * The check_bounds block can come in from handle_tt_uint or
       * handle_tt_sint, so the icmps it uses are phis
       */
      LLVMPositionBuilderAtEnd(ctxt->builder, check_bounds);
      H_LLVM_COMPUTE_NAME(ir_lower_icmp, p);
      lower_icmp = LLVMBuildPhi(ctxt->builder,
          LLVMTypeOf(lower_uint_icmp), ir_lower_icmp_name);
      lower_icmp_incoming_blocks[0] = handle_tt_uint;
      lower_icmp_incoming_values[0] = lower_uint_icmp;
      lower_icmp_incoming_blocks[1] = handle_tt_sint;
      lower_icmp_incoming_values[1] = lower_sint_icmp;
      LLVMAddIncoming(lower_icmp,
                      lower_icmp_incoming_values,
                      lower_icmp_incoming_blocks,
                      2);
      H_LLVM_COMPUTE_NAME(ir_upper_icmp, p);
      upper_icmp = LLVMBuildPhi(ctxt->builder,
          LLVMTypeOf(upper_uint_icmp), ir_upper_icmp_name);
      upper_icmp_incoming_blocks[0] = handle_tt_uint;
      upper_icmp_incoming_values[0] = upper_uint_icmp;
      upper_icmp_incoming_blocks[1] = handle_tt_sint;
      upper_icmp_incoming_values[1] = upper_sint_icmp;
      LLVMAddIncoming(upper_icmp,
                      upper_icmp_incoming_values,
                      upper_icmp_incoming_blocks,
                      2);
      /* Did we fail the lower bound? */
      LLVMBuildCondBr(ctxt->builder, lower_icmp, fail, check_upper_bound);

      /* Okay, now we check the upper one */
      LLVMPositionBuilderAtEnd(ctxt->builder, check_upper_bound);
      LLVMBuildCondBr(ctxt->builder, upper_icmp, fail, success);

      /*
       * From success, branch to end (and the phi for rv will know to
       * use ret)
       */
      LLVMPositionBuilderAtEnd(ctxt->builder, success);
      LLVMBuildBr(ctxt->builder, end);

      /*
       * From fail, branch to end (and the phi for rv will know to
       * use NULL)
       */
      LLVMPositionBuilderAtEnd(ctxt->builder, fail);
      LLVMBuildBr(ctxt->builder, end);

      /*
       * In end, phi for rv
       */
      LLVMPositionBuilderAtEnd(ctxt->builder, end);
      H_LLVM_COMPUTE_NAME(ir_rv, p);
      rv = LLVMBuildPhi(ctxt->builder, ctxt->llvm_parseresultptr, ir_rv_name);
      rv_phi_incoming_blocks[0] = success;
      rv_phi_incoming_values[0] = inner_res;
      rv_phi_incoming_blocks[1] = fail;
      rv_phi_incoming_values[1] = LLVMConstNull(ctxt->llvm_parseresultptr);
      LLVMAddIncoming(rv, rv_phi_incoming_values, rv_phi_incoming_blocks, 2);
      *res = rv;

      /* Pass it back out */
      *res = rv;
    } else {
      status = false;
    }
  } else {
    status = false;
  }

  /* Free names */
  H_LLVM_FREE_NAME(ir_entry);
  H_LLVM_FREE_NAME(ir_check_res);
  H_LLVM_FREE_NAME(ir_check_res_ast);
  H_LLVM_FREE_NAME(ir_check_token_type);
  H_LLVM_FREE_NAME(ir_check_tt_uint);
  H_LLVM_FREE_NAME(ir_check_tt_sint);
  H_LLVM_FREE_NAME(ir_handle_tt_uint);
  H_LLVM_FREE_NAME(ir_handle_tt_sint);
  H_LLVM_FREE_NAME(ir_check_bounds);
  H_LLVM_FREE_NAME(ir_check_upper_bound);
  H_LLVM_FREE_NAME(ir_success);
  H_LLVM_FREE_NAME(ir_fail);
  H_LLVM_FREE_NAME(ir_end);
  H_LLVM_FREE_NAME(ir_inner_res_icmp);
  H_LLVM_FREE_NAME(ir_ret_ast_ptr);
  H_LLVM_FREE_NAME(ir_ret_ast);
  H_LLVM_FREE_NAME(ir_ret_ast_icmp);
  H_LLVM_FREE_NAME(ir_token_type_ptr);
  H_LLVM_FREE_NAME(ir_token_type);
  H_LLVM_FREE_NAME(ir_token_data_ptr);
  H_LLVM_FREE_NAME(ir_token_data);
  H_LLVM_FREE_NAME(ir_tt_uint_icmp);
  H_LLVM_FREE_NAME(ir_tt_sint_icmp);
  H_LLVM_FREE_NAME(ir_lower_uint_icmp);
  H_LLVM_FREE_NAME(ir_lower_sint_icmp);
  H_LLVM_FREE_NAME(ir_lower_icmp);
  H_LLVM_FREE_NAME(ir_upper_uint_icmp);
  H_LLVM_FREE_NAME(ir_upper_sint_icmp);
  H_LLVM_FREE_NAME(ir_upper_icmp);
  H_LLVM_FREE_NAME(ir_rv);

  return status;
}

#endif

static const HParserVtable int_range_vt = {
  .parse = parse_int_range,
  .isValidRegular = h_true,
  .isValidCF = h_true,
  .desugar = desugar_int_range,
  .compile_to_rvm = ir_ctrvm,
#ifdef HAMMER_LLVM_BACKEND
  .llvm = ir_llvm,
#endif
  .higher = false,
};

HParser* h_int_range(const HParser *p, const int64_t lower, const int64_t upper) {
  return h_int_range__m(&system_allocator, p, lower, upper);
}
HParser* h_int_range__m(HAllocator* mm__, const HParser *p, const int64_t lower, const int64_t upper) {
  // p must be an integer parser, which means it's using parse_bits
  // TODO: re-add this check
  //assert_message(p->vtable == &bits_vt, "int_range requires an integer parser"); 

  // and regardless, the bounds need to fit in the parser in question
  // TODO: check this as well.

  HRange *r_env = h_new(HRange, 1);
  r_env->p = p;
  r_env->lower = lower;
  r_env->upper = upper;
  return h_new_parser(mm__, &int_range_vt, r_env);
}
